// $Id: nntp.cpp,v 1.5 2002/04/23 18:21:35 jjlupa Exp $
//
// ypost (Yet Another Poster)
// This code is copyright John Ridley, 2000. <john.ridley@gt.org>  To the
//  extent that it is released, it is released under the GNU General Public License
//  v2 or greater. The author retains all rights with respect to this code.
//
// If you have a desire to use this code outside of the GPL, drop me an email and
//  I would be happy to work with you under an alternate licence.  If you just want
//  to write code, then I would probably be happy to release it to you as under
//  the perl artistic license.

// nntp.cpp
// very simple nntp class - all it does is post articles

namespace std
    {}
using namespace std;

#ifdef WIN32 // God can this get annoying...
#   pragma warning(disable:4786)
#endif

#include "ypost.h"
#include "nntp.h"
#include "settings.h"
#include "independent.h"
#include "common.h"

#include <iostream>
#include <assert.h>

#ifdef WIN32
#   include <winsock2.h>
#else
#   include <stdio.h>
#   include <sys/time.h>
#   include <unistd.h>
#endif


namespace
{
    string int2str( const int i )
    {
        char buf[20]; // MAX_INT == 2147483647
        memset( buf, 0, sizeof( buf ) );
        sprintf( buf, "%i", i );
        return string( buf );
    }

    const long progressBarLength_ = 50;
    static long s_lastpct = 0;
    void progress( long done, long total )
    {
        if ( userSettings::obj().verbosity < 2 )
            return;

        long pct = ( done * 100) / total;
        if (0 > pct)
        {
            pct = 0;
            s_lastpct = -1;
        }
        if (pct > 100)
            pct = 100;

        if (pct == s_lastpct)
            return;
        s_lastpct = pct;

        cerr << "\r|";
        int num = (progressBarLength_ * pct) / 100;
        int x;
        for (x=0; x < progressBarLength_; x++)
        {
            if (x < num)
                cerr << "*";
            else
                cerr << "_";
        }
        cerr << "| " << pct << "%  ";

        if ( done == total )
            cerr << endl;
    }
}

nntp::nntp(const char *server, int port) :
        m_serverName(server),
        m_port(port),
        m_pSock(0),
        m_bConnected(false)
{}

nntp::~nntp()
{
    disconnect();
}

void nntp::setAuthInfo(const char *userid, const char *passwd)
{
    m_userid = userid;
    m_passwd = passwd;
}

bool nntp::getResultCode()
{
    if ( NULL == m_pSock )
        return false;

    string read_buf;
    bool result = m_pSock->readline( read_buf, 512 );
    m_lastResult = read_buf;

    if ( userSettings::obj().verbosity > 3 )
        cerr << "result:" << m_lastResult << endl;

    // This best be true
    if ( result )
        assert( m_lastResult.size() );

    return result;
}


/* ***********************************************************
establish the connection and look for a "200" return code
send auth info if we have any
*/
bool nntp::connect()
{
    if (m_bConnected)
        return true;

    if (userSettings::obj().verbosity > 3)
        cerr << "Connecting to " << m_serverName << ":" << m_port << endl;

    m_pSock = new sock::socket();
    bool connected = m_pSock->connect( sock::sockaddr( m_serverName, int2str( m_port ) ) );
    if ( ! getResultCode() )
        throw Ypost_Exception(Ypost_Exception::_NONFATAL_, std::string("Unable to create IP socket"));

    if (m_lastResult[0] == '2')  // 2xx we're happy with
    {
        // send auth info if available
        if (!m_userid.empty())
        {
            // >AUTHINFO USER myuserid
            // <381 PASS required
            // >AUTHINFO PASS mypassword
            // <281 Ok
            if (userSettings::obj().verbosity > 3)
                cerr << "Sending userid " << m_userid << endl;

            string auth( "AUTHINFO USER " );
            auth += m_userid;
            m_pSock->writeline( auth );

            if (   ( ! getResultCode() ) || (m_lastResult.substr(0,3) != std::string("381")))
                throw Ypost_Exception( Ypost_Exception::_NONFATAL_, std::string("Bad result while sending auth info:") + m_lastResult );

            if (userSettings::obj().verbosity > 3)
                cerr << "Sending password" << endl;
            m_pSock->writeline("AUTHINFO PASS ", sock::socket::null_terminator );
            m_pSock->writeline(m_passwd.c_str());
            if (   (!getResultCode())
                    || (m_lastResult.substr(0,3) != std::string("281")))
                throw Ypost_Exception(Ypost_Exception::_NONFATAL_,
                                      std::string("Bad result while sending auth info")
                                      + m_lastResult);
        }
    }

    if (userSettings::obj().verbosity > 3)
        cerr << "Logged in." << endl;

    if (userSettings::obj().defined("throttle"))
    {
        m_pSock->write_throttle( atol(userSettings::obj().getValue("throttle").c_str()) );
    }
    m_bConnected = true;
    return m_bConnected;
}

bool nntp::disconnect()
{
    if (m_bConnected)
    {
        if (userSettings::obj().verbosity > 3)
            cerr << "Logging out of server" << endl;
        m_pSock->writeline("QUIT");
        getResultCode();
        // we don't really care what the result is, we're outta here
        m_bConnected = 0;
    }
    if (0 != m_pSock)
    {
        delete(m_pSock); // this will terminate the connection
        m_pSock = 0;
    }

    return true;
}



/* **** here's a simple sample ****
POST
340 send article to be posted.
From: Traalfaz@hotmail.com
Newsgroups: alt.test
Subject: test using telnet
 
 test posted with telnet
 .
 240 ok
*/
bool nntp::postArticle(const char *group,
                       const char *from,
                       const char *subject,
                       const char *xheaders,
                       const char *body)
{
    m_bytesPosted = 0;
    if (!m_bConnected)
        connect();

    if (userSettings::obj().verbosity > 3)
        cerr << "Requesting begin POST" << endl;
    m_pSock->writeline("POST");
    getResultCode();
    if (m_lastResult[0] != '3')  // we must get 3xx back
        throw Ypost_Exception(Ypost_Exception::_NONFATAL_,
                              std::string("Bad result when beginning POST: ") + m_lastResult);

    std::string headers;
    headers = std::string("From: ") +
              std::string(from) +
              std::string("\r\nNewsgroups: ") +
              std::string(group) +
              std::string("\r\nSubject: ") +
              std::string(subject) +
              std::string("\r\nX-Newsposter: " YPOST_VER_); // always add this

    if (NULL != xheaders)
        headers += std::string("\r\n") + std::string(xheaders);

    // Needs an extra linefeed
    headers += std::string("\r\n");

    if (userSettings::obj().verbosity > 4)
        cerr << "Sending headers:" << endl << headers << endl;

    m_bytesPosted += headers.size();

    m_pSock->writeline(headers.c_str());

    if (userSettings::obj().verbosity > 4)
        cerr << "Sending body" << endl;

    m_bytesPosted += strlen(body);

    s_lastpct = 0;
    timeval t1;
    independent::get_time( t1 );
    m_pSock->set_progress_callback( progress );
    m_pSock->writeline(body); // with progress bar
    m_pSock->set_progress_callback( NULL );
    timeval t2;
    independent::get_time( t2 );
    m_pSock->writeline("\r\n.");

    if ( userSettings::obj().verbosity > 2 )
    {
        const unsigned int time = common::timeval_diff( t2, t1 );
        if ( time )
        {
            const unsigned int bytes = strlen( body );
            cerr << "Posted body at " << ( bytes*1000000.0 / time ) << " bytes per second...(" << bytes << " bytes in " << time/1000000.0 << " seconds)."  <<  endl;
        }
        else
            cerr << "ERROR: Bad time value while posting body. Can't determine write rate." << endl;
    }

    if (userSettings::obj().verbosity > 2)
        cerr << "Awaiting confirmation of article from server" << endl;

    getResultCode();

    if (m_lastResult[0] == '2')
    {
        if (userSettings::obj().verbosity == 3) // above 3 already prints
            cerr << m_lastResult << endl;
        return true;
    }
    throw Ypost_Exception(Ypost_Exception::_NONFATAL_,
                          std::string("Bad result at end of POST"));
}
