// $Id: settings.cpp,v 1.5 2002/04/23 18:21:00 jjlupa Exp $
//
// ypost (Yet Another Poster)
// This code is copyright John Ridley, 2002. <john.ridley@gt.org>  To the
//  extent that it is released, it is released under the GNU General Public License
//  v2 or greater. The author retains all rights with respect to this code.
//
// If you have a desire to use this code outside of the GPL, drop me an email and
//  I would be happy to work with you under an alternate licence.  If you just want
//  to write code, then I would probably be happy to release it to you as under
//  the perl artistic license.

// settings.cpp
// Global object that contains user's default and selected options

namespace std
    {}
using namespace std;

#ifdef WIN32 // God can this get annoying...
#   pragma warning(disable:4786)
#endif

#include "ypost.h"
#include "settings.h"
#include "independent.h"

#include <iostream>
#include <string>

#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>

#ifdef WIN32
#   define snprintf _snprintf
#else
#   include <stdio.h>
#   include <unistd.h>
#endif

// ***********************************************
// singleton access
// ***********************************************
userSettings &userSettings::obj()
{
    static userSettings me;
    return me;
}

// ***********************************************
// constructor
// ***********************************************
userSettings::userSettings() :
        m_invalidArgFound(false),
        m_writeDefaults(false),
        m_binEncoding(FileObj::_bin_yenc_),
        verbosity(1)
{
    defaults();
    readSettingsFile();
}

void userSettings::defaults()
{
    m_values[string("port")] = string("119");
    m_values[string("lines")] = string("4500");
}

// ***********************************************
// parse command line args
// ***********************************************
void userSettings::parseArgs(int argc, char *argv[])
{
    int idx(1);
    while (idx < argc)
    {
        idx += parseOneArg(argc-idx,&(argv[idx]));
    }
    if (m_writeDefaults)
    {
        writeSettingsFile();
        m_writeDefaults = false;
    }
    validateArgs();
    // if filelist is empty, print usage and exit
    printJobSummary();
}

// ***********************************************
// print what we're going to do at the end of parsing
// ***********************************************
void userSettings::printJobSummary()
{
    if (verbosity < 1)
        return;
    printVersion();
    cerr
    << "From       : " << m_values[string("from")] << endl
    << "Subject    : " << m_values[string("subject")] << endl
    << "Newsgroups : " << m_values[string("newsgroups")] << endl
    << "News server: " << m_values[string("hostname")] << endl
    << m_values[string("xheaders")] << endl;

    if (verbosity > 1)
    {
        long TotalSize(0);

        cerr << "Files:" << endl;
        int NumFiles = userSettings::obj().numFiles();
        for (int filenum=0; filenum < NumFiles; filenum++ )
        {
            long bytes = userSettings::obj().getFileObj(filenum).getBytes();
            cerr << userSettings::obj().getFileObj(filenum).getFilename()
            << "\t(" << bytes << " bytes)\t";
            TotalSize += bytes;
            switch (userSettings::obj().getFileObj(filenum).getEncodeMethod())
            {
            case FileObj::_text_     :
                cerr << " (as text)";
                break;
            case FileObj::_bin_uu_   :
                cerr << " (uuencode)";
                break;
            case FileObj::_bin_yenc_ :
                cerr << " (yencode)";
                break;
            default                  :
                cerr << " (UNKNOWN ENCODING)";
                break;
            }
            cerr << endl;
        }
        cerr
        << "Total: " << TotalSize << " bytes" << endl << endl;
    }
    cerr << "=============================================================" << endl;
}

// ***********************************************
// given the current head of the args, parse one thing from it
// ***********************************************
int userSettings::parseOneArg(int argc, char *argv[])
{
    int numArgsEaten = 1; // always at least one

    // get arg
    char *arg = argv[0];

    // if it starts with a '-' it's an arg, else it's a file
    if (*arg != '-')
    {
        // if there's a colon, then it's got parts, else whole file
        char *tptr = strchr(arg,':');
        if (NULL != tptr)
        {
            *(tptr++) = 0; /* chop at the end of the filename and parse after the colon */
            // it's a file; push it on the stack
            FileObj fo(arg,m_binEncoding);
            parseparts(tptr, fo);
            m_files.push_back(fo);
        }
        else
        {

            // If the file has an * in it, we need to expand it for win32. The unix shell expands
            // files before they get into argv, win32 does not...
#ifdef WIN32
            vector<string> filenames;
            win32::expand_fns( arg, filenames );
            for ( vector<string>::const_iterator ii = filenames.begin() ; ii != filenames.end() ; ++ii )
                m_files.push_back( FileObj( (*ii).c_str(), m_binEncoding ) );
#else

            m_files.push_back(FileObj(arg, m_binEncoding));
#endif

        }

    }
    else
    {

        // it's an argument;
        switch (arg[1])
        {
        case '?':
            printUsage();
            exit(0);

        case 'n':
            recordExtraArg(argc, argv, "newsgroups");
            numArgsEaten++;
            break;

        case 's':
            recordExtraArg(argc, argv, "subject");
            numArgsEaten++;
            break;

        case 'f':
            recordExtraArg(argc, argv, "from");
            numArgsEaten++;
            break;

        case 'h':
            recordExtraArg(argc, argv, "hostname");
            numArgsEaten++;
            break;

        case 'p':
            if (arg[2] == 'o')
            {
                recordExtraArg(argc, argv, "port");
            }
            else
            {
                recordExtraArg(argc, argv, "password");
            }
            numArgsEaten++;
            break;

        case 'u':
            if (arg[2] == 'u')
                m_binEncoding = FileObj::_bin_uu_;
            else
            {
                recordExtraArg(argc, argv, "userid");
                numArgsEaten++;
            }
            break;

        case 'y':
            m_binEncoding = FileObj::_bin_yenc_;
            break;

        case 'l':
            recordExtraArg(argc, argv, "lines");
            numArgsEaten++;
            break;

        case 't':
            if (arg[2] == 'h')
                recordExtraArg(argc, argv, "throttle");
            else
            {
                if (argc < 2)
                {
                    printUsage();
                    throw Ypost_Exception(Ypost_Exception::_FATAL_,
                                          string("invalid arguments"));
                }
                // it's a file; push it on the stack
                m_files.push_back(FileObj(argv[1],FileObj::_text_));
            }
            numArgsEaten++;
            break;

        case 'r':
            recordExtraArg(argc, argv, "retry");
            numArgsEaten++;
            break;

        case 'd':
            m_writeDefaults = true; // write defaults after parsing
            break;

        case 'v':
            verbosity = atoi(argv[1]);
            numArgsEaten++;
            break;

        default:
            cerr << "Invalid argument: " << argv[0] << endl;
            m_invalidArgFound = true;
        }
    }
    return numArgsEaten;
}


/* helper for parts parser; comes up with either numbers or delimiters */
/* returns pointer to next char in string */
static const char *gettoken(const char *str, char *buf)
{
    if (*str >= '0' && *str <= '9')
    {
        for (;*str >= '0' && *str <= '9'; str++)
        {
            *(buf++) = *str;
        }
        *buf = 0;
        return str;
    }
    switch (*str)
    {
    case ',':
    case '-':
        *(buf++) = *(str++);
        *buf = 0;
        break;
    default:
        *buf = 0;
        if (*str)
            str++;
        break;
    }
    return str;
}

/* look for a parts list at the end of a filename */
/* filename:1-4,8,11 */
/* If no parts list, return null */
bool userSettings::parseparts(const char *filearg, FileObj &fo)
{
    char numbuf[32];
    const char * tptr = filearg;
    int seqstart, thisnum, swap, ii;

    seqstart = -1;
    thisnum = -1;
    do
    {
        tptr = gettoken(tptr, numbuf);
        if (*numbuf >= '0' && *numbuf <= '9')
        {
            thisnum = atoi(numbuf);
        }
        else
        {
            if (*numbuf == '-')
                seqstart = thisnum;
            else
            {
                if (seqstart != -1)
                {
                    if (thisnum < seqstart)
                    {
                        swap = thisnum;
                        thisnum = seqstart;
                        seqstart = swap;
                    }
                    for (ii = seqstart; ii <= thisnum; ii++)
                        fo.m_partsToPost.push_back(ii);
                    seqstart = -1;
                }
                else
                {
                    if (-1 != thisnum)
                    {
                        /* inserting just the one number */
                        fo.m_partsToPost.push_back(thisnum);
                    }
                }
            }
        }
    }
    while ((*numbuf) && (thisnum > 0));
    return true;
}

// check for required values
bool userSettings::validateArgs()
{
    char *reqVals[] = {"from","subject","newsgroups","hostname",NULL};
    string strError;

    for (char **key = reqVals; *key; key++)
    {
        if (!defined(*key))
        {
            strError +=
                string("You must provide a ") + string(*key)
                + string(" parameter.\n");
        }
    }

    // check for existance of all files
    int NumFiles = numFiles();
    if (0 == NumFiles)
    {
        strError += string("No files to post!\n");
    }
    int filenum;
    for (filenum=0; filenum < NumFiles; filenum++ )
    {
        struct stat Stats;
        if (0 > stat(getFileObj(filenum).getFilename().c_str(), &Stats))
        {
            strError +=
                string("File ") + getFileObj(filenum).getFilename()
                + string(" not found!\n");
        }
        else
        {
            getFileObj(filenum).setBytes(Stats.st_size);
        }
    }
    if (!strError.empty())
    {
        printUsage();
        throw Ypost_Exception(Ypost_Exception::_FATAL_,strError);
    }
    return true;
}

// returns true if there's a value for the given key
bool userSettings::defined(const char *pszKey)
{
    return (m_values.find(string(pszKey)) != m_values.end());
}

string userSettings::getValue(const char *pszKey)
{
    return m_values[string(pszKey)];
}

string userSettings::getValue( const char * pszKey, const char * def )
{
    assert( pszKey && def );
    if ( ! defined( pszKey ) )
        m_values.insert( make_pair( string( pszKey ), string( def ) ) );
    return m_values[string(pszKey)];
}

int userSettings::numFiles()
{
    return m_files.size();
}

FileObj &userSettings::getFileObj(int idx)
{
    // todo: sanity checking
    return m_files[idx];
}


static string getrcfilename(const char *filename)
{
    string strFilename;
    if (0 == filename)
    {
        char *homedir = getenv("HOME");
        if (homedir != NULL)
            strFilename = homedir;
        else
        {
#ifdef WIN32
            strFilename = "c:\\";
#else

            strFilename = "~";
#endif

        }
        strFilename += independent::slash();
        strFilename += ".ypostrc";
    }
    else
        strFilename = filename;
    return strFilename;
}

// serialization
// format: key\tvalue\tvalue\tvalue (\t after 1st one indicates c/r in value)
// special key(s):
//   encoding:uu or yenc
//   verbosity:(number)
void userSettings::readSettingsFile(const char *filename /* = 0 */)
{
    string strFilename(getrcfilename(filename));
    FILE *rcfile;
    rcfile = fopen(strFilename.c_str(), "r");
    if (NULL == rcfile)
        return; // this is OK

    char buf[256];
    while (fgets(buf,256,rcfile))
    {
        char *tptr;
        // whack the newline at the end
        tptr = strchr(buf,'\n');
        if (NULL != tptr)
            *tptr = 0;
        tptr = strchr(buf,'\t');
        if (NULL == tptr)
	{
	    // We didn't find a tab.  Now lets be super cool and see if we have something
	    // of the form ASDF     FDAS where those spaces are spaces. I'm only going to
	    // validate things which have 2 tokens...
	    const char * sep = " \t";
	    char * token = strtok( buf, sep );
	    vector<string> toks;
	    while( token ) 
	    {
		toks.push_back( string( token ) );
		token = strtok( NULL, sep );
	    }
	    if ( toks.size() == 2 )
	    {
		m_values[(toks[0])]=toks[1];
	    }
	    continue;
	}
        *(tptr++) = 0;
        string key(buf);
        char *valptr = tptr;
        while (tptr = strchr(tptr,'\t'))
            *tptr = '\n';
        string value(valptr);

        if (key == string("encoding"))
        {
            if (value == string("uu"))
                m_binEncoding = FileObj::_bin_uu_;
            else
                m_binEncoding = FileObj::_bin_yenc_;
        }
        else
        {
            if (key == string("verbosity"))
            {
                verbosity = atoi(value.c_str());
            }
            else
            {
                m_values[key] = value;
            }
        }
    }
    fclose(rcfile);
}


// this saves the current settings to a settings file.

void userSettings::writeSettingsFile(const char *filename /* = 0 */)
{
    string strFilename(getrcfilename(filename));
    FILE *rcfile;
    rcfile = fopen(strFilename.c_str(), "w");
    if (NULL == rcfile)
        return;  // I guess this is OK, but maybe not.

    map<string, string>::iterator ii = m_values.begin();
    while (ii != m_values.end())
    {
        string line((*ii).first);
        line += "\t";
        line += (*ii).second;
        int pos = line.find_first_of('\n');
        while (-1 != pos)
        {
            line[pos] = '\t';
            pos = line.find_first_of('\n');
        }
        line += '\n';
        fputs(line.c_str(),rcfile);
        ii++;
    }
    string encoding("encoding\t");
    if (m_binEncoding == FileObj::_bin_uu_)
        encoding += "uu\n";
    else
        encoding += "yenc\n";
    fputs(encoding.c_str(), rcfile);
    char chbuf[64];
    snprintf(chbuf,sizeof(chbuf), "verbosity\t%d\n",verbosity);
    fputs(chbuf, rcfile);
    fclose(rcfile);
}




// ***********************************************
// helper for parseOneArg
// if arg is empty ("") then delete this map item
// ***********************************************
void userSettings::recordExtraArg(int argc, char *argv[],
                                  const char *cmdname,
                                  bool xheaders /* = false */)
{
    if (argc < 2)
    {
        throw Ypost_Exception(Ypost_Exception::_FATAL_,
                              string("invalid arguments"));
    }
    string value(argv[1]);
    if (xheaders)
    {
        if (value.empty())
        {
            map<string, string>::iterator ii =
                m_values.find(string("xheaders"));
            m_values.erase(ii);
        }
        else
        {
            string tmpstr(cmdname);
            tmpstr += string(argv[1]);
            if (!(m_values[string("xheaders")].empty()))
                m_values[string("xheaders")] += string("\n");
            m_values[string("xheaders")] += tmpstr;
        }
    }
    else
    {
        string key(cmdname);
        if (value.empty())
        {
            map<string, string>::iterator ii = m_values.find(key);
            if (ii != m_values.end())
                m_values.erase(ii);
        }
        else
        {
            m_values[key] = value;
        }
    }
}



// ***********************************************
// prints the version (duh)
// ***********************************************
void userSettings::printVersion()
{
    cerr << YPOST_VER_ << endl;
}

// ***********************************************
// print help (duh again)
// ***********************************************
void userSettings::printUsage()
{
    printVersion();
    cerr
    << "Usage:  ypost [options] file1[:parts] ..." << endl
    << "      valid options:" << endl
    << "        -from email-addr" << endl
    << "        -newsgroups groupnames - comma separated" << endl
    << "        -Subject SUBJ - subject for all your posts." << endl
    << "        -hostname (name) - ip or name of posting host" << endl
    << "        -port (portnum)  - port on server to post to" << endl
    << "        -userid (userid) - user id if authentication is required" << endl
    << "        -passwd (password) - passwd if auth required" << endl
    << "        -lines (numlines) - max lines per post" << endl
    << "        -throttle (Bps) - max bytes per sec to post" << endl
    << "        -text filename - post as text file" << endl
    << "        -uuencode - use uu encoding for subsequent files" << endl
    << "        -yenc     - use yenc encoding for subsequent files" << endl
    << "        -v xxx    - set verbosity level to xxx" << endl
    << "           0 = quiet (errors only)" << endl
    << "           1 = normal, 2 = progress bars" << endl
    << "           3 = basic debugging, 4 = more debugging, 5 = very noisy" << endl
    << "        -defaults - write current settings to ~/.ypostrc" << endl
    << "        -? print this help" << endl
    << endl;
}



// ******************************************************
// FileObj implementation
// ******************************************************
FileObj::FileObj(const char *pszFilename, encodeMethod method) :
        m_filename(pszFilename),
        m_method(method)
{}

bool FileObj::postThisPart(int part)
{
    if (m_partsToPost.empty())
        return true;
    int i;
    for (i=0; i < m_partsToPost.size(); i++)
    {
        if (m_partsToPost[i] == part)
            return true;
    }
    return false;
}
