// $Id: socket.cpp,v 1.6 2002/04/26 15:18:13 jjlupa Exp $
//
// ypost (Yet Another Poster)
// This code is copyright Jonathan Lupa, 2000. <jjlupa@jamdata.net>  To the
//  extent that it is released, it is released under the GNU General Public License
//  v2 or greater. The author retains all rights with respect to this code.
//
// If you have a desire to use this code outside of the GPL, drop me an email and
//  I would be happy to work with you under an alternate licence.  If you just want
//  to write code, then I would probably be happy to release it to you as under
//  the perl artistic license.

// sockets.cpp
// Simple sockets wrapping

namespace std
    {}
using namespace std;

#ifdef WIN32 // God can this get annoying...
#   pragma warning(disable:4786)
#endif

#include "socket.h"
#include "independent.h"
#include "settings.h"

#include <assert.h>
#include <stdio.h>
#include <errno.h>

#include <iostream>
#include <string>
#include <stdexcept>
#include <fstream>
#include <deque>

#ifdef WIN32
#   include <winsock2.h>
#   include <io.h>
typedef int socklen_t;
#else
#   include <netinet/in.h>
#   include <netinet/tcp.h>
#   include <sys/socket.h>
#   include <arpa/inet.h>
#   include <sys/time.h>
#   include <netdb.h>
#   include <unistd.h>
#define SOCKET_ERROR (-1)
#endif

namespace
{
    //
    // str2ushort() :
    //
    unsigned short str2ushort( const string & s )
    {
        // Max ushort is 65535, chopping from the int should
        // be good enough
        return atoi( s.c_str() );
    }

    //
    // int2str() :
    //
    string int2str( const int i )
    {
        // Max int is 2147483647
        char buf[12];
        sprintf( buf, "%d", i );
        return string( buf );
    }

    //
    // CHoldIntUntilDestruction() : Holds a reference to the passed in int and zeroes it out for the
    //   lifetime of the object.  This just helps me keep throttling out of places it doesn't belong.
    //
    class CHoldIntUntilDestruction
    {
    public:
        CHoldIntUntilDestruction( unsigned int & in ) : m_hold( in ), m_ref( in )
        {
            m_ref = 0;
        }
        ~CHoldIntUntilDestruction()
        {
            m_ref = m_hold;
        }
    private:
        const unsigned int m_hold;
        unsigned int & m_ref;
    };

    //
    // simple delta function for timevals
    //
    unsigned long tval_diff( const timeval & t1, const timeval & t2 )
    {
        assert( t1.tv_sec > t2.tv_sec || ( t1.tv_sec == t2.tv_sec && t1.tv_usec >= t2.tv_usec ) );
        return ( ( ( t1.tv_sec * 1000000 ) + t1.tv_usec ) - ( ( t2.tv_sec * 1000000 ) + t2.tv_usec ) );
    }

    //
    // simple_stats :  The Throttle managed in the send loop. This just helps collect the data.
    //
    // If this gets too thick, we could have performance issues.
#define STAT_QUEUE_DEPTH 10
    class simple_stats
    {
    public:
        // addnode() :  given a new node value, computes the windowed rate... side effects the queue depth clipping
        unsigned long addnode( unsigned long bytes, timeval t )
        {
            if ( bytes == 0 && m_history.size() > 1 )
            {
                // If we are just spinning on a waits and our queue has data, lets just keep bumping
                // the last time record... this way we lose no data...
                m_history.back().t = t;
            }
            else
            {
                if ( m_history.size() >= STAT_QUEUE_DEPTH )
                    m_history.erase( m_history.begin() );

                node n;
                n.bytes = bytes;
                n.t = t;
                m_history.push_back( n );
            }
            return compute_rate();
        }

        unsigned long compute_rate()
        {
            timeval start = m_history.front().t;
            timeval end   = m_history.back().t;

            unsigned long bytes = 0;
            for( deque<node>::iterator ii = m_history.begin() ; ii != m_history.end() ; ++ii )
                bytes += (*ii).bytes;

            const unsigned long tot_usec = tval_diff( end, start );
            if ( tot_usec == 0 )
                return 0;

            const unsigned long ret = (unsigned long)( (float)bytes / (float)tot_usec * 1000000.0 ) ;
            return ret;
        }
    private:
        struct node
        {
            unsigned long bytes;
            timeval       t;
        };

        deque<node> m_history;
    };
}

namespace sock
{
    ////////////////////////////////////////////////////////////
    //
    // sockaddr_imp
    //
    class sockaddr_imp
    {
        // INTERFACE
    public:

        //
        // set_address() : Workhorse that interprets the input and turns it into a workable in_addr_whatever
        //
        bool set_address( const std::string & address_name, const std::string & service_name )
        {
            m_valid = false;

            // Save these off in case we want them later....
            m_address = address_name;
            m_service = service_name;

            memset( &m_addr_v4, 0, sizeof( m_addr_v4 ) );

            // setup the server address....
            if ( m_address == "ANY" )
            {
                const long addr_to_use = htonl( INADDR_ANY );
                memcpy( &m_addr_v4.sin_addr, &addr_to_use, sizeof( addr_to_use ) );
            }
            else
            {
                unsigned long int_address = inet_addr( address_name.c_str() );
                if ( int_address != INADDR_NONE )
                {
                    // The address passed in is a dotted quad.
                    memcpy( &m_addr_v4.sin_addr, &int_address, sizeof( int_address ) );
                }
                else
                {
                    hostent * he = gethostbyname( address_name.c_str() );
                    // If we couldn't look up the name, lets just punt.
                    if ( ! he )
                        return false;
                    memcpy( &m_addr_v4.sin_addr, he->h_addr, he->h_length );
                }
            }

            // Setup the easy stuff...
            m_addr_v4.sin_family = AF_INET;
            m_addr_v4.sin_port = htons( str2ushort( service_name ) );

            m_valid = true;

            return true;
        }

        //
        // set_address() : This version of set address is to create an address out of preexisting internal data.
        //
        bool set_address( const ::sockaddr * addr_in )
        {
            if ( ! addr_in )
            {
                assert( false );
                return false;
            }

            if ( addr_in->sa_family == AF_INET )
            {
                m_address = string( inet_ntoa( reinterpret_cast<const sockaddr_in *>( addr_in )->sin_addr ) );
                m_service = int2str( ntohs( reinterpret_cast<const sockaddr_in *>( addr_in )->sin_port ) );
            }

            // else not supported address family
            return false;
        }

        //
        // construction
        //
        sockaddr_imp() : m_valid( false )
        {}
        sockaddr_imp( const std::string & address_name, const std::string & service_name ) :
                m_valid( false )
        {
            set_address( address_name, service_name );
        }

        //
        // is_valid() :
        //
        bool is_valid() const
        {
            return m_valid;
        }

        //
        // data access
        //
        const ::sockaddr * get_address() const
        {
            if ( ! is_valid() )
                return NULL;
            return reinterpret_cast<const ::sockaddr *>( &m_addr_v4 );
        }
        size_t             get_address_size() const
        {
            return sizeof( m_addr_v4 );
        }

        std::string get_verbose_address() const
        {
            return m_address + ":" + m_service;
        }

        // IMPLEMENTATION
    private:
        string m_address;
        string m_service;

        sockaddr_in m_addr_v4;
        bool   m_valid;
    };



    ////////////////////////////////////////////////////////////
    //
    // sockaddr
    //
    // This all adds up to a thin wrapper on the implementation.
    sockaddr::sockaddr() : m_pimp( NULL )
    {
        m_pimp = new sockaddr_imp;
    }
    sockaddr::sockaddr( const std::string & address_name, const std::string & service_name ) : m_pimp( NULL )
    {
        m_pimp = new sockaddr_imp( address_name, service_name );
    }
    bool sockaddr::set_address( const std::string & address_name, const std::string & service_name )
    {
        assert( m_pimp );
        return m_pimp->set_address( address_name, service_name );
    }
    bool sockaddr::set_address( const ::sockaddr * addr_in )
    {
        assert( m_pimp );
        return m_pimp->set_address( addr_in );
    }
    bool sockaddr::is_valid() const
    {
        assert( m_pimp );
        return m_pimp->is_valid();
    }
    const ::sockaddr * sockaddr::get_address() const
    {
        assert( m_pimp );
        return m_pimp->get_address();
    }
    size_t sockaddr::get_address_size() const
    {
        assert( m_pimp );
        return m_pimp->get_address_size();
    }
    std::string sockaddr::get_verbose_address() const
    {
        assert( m_pimp );
        return m_pimp->get_verbose_address();
    }


    ////////////////////////////////////////////////////////////
    //
    // socket
    //
    const std::string socket::null_terminator("");
    //
    // ctor/dtor
    //
    socket::socket() :
            m_sockfd( -1 ),
            m_readRate( 0 ),
            m_writeRate( 0 ),
            m_cb( NULL )
    {}

    socket::~socket()
    {
        if ( is_connected() )
            close();
    }

    //
    // attach() : take ownership of a socket
    //
    void socket::attach( int fd )
    {
        m_sockfd = fd;
    }

    //
    // readbuf() :
    //
    bool socket::readbuf( void * buf, const size_t bufsz, unsigned int wait_ms )
    {
        assert( is_connected() &&  buf && bufsz > 0 );

        // Wait for input...
        if ( wait_ms )
        {
            fd_set rfds;
            FD_ZERO( &rfds );
            FD_SET( m_sockfd, &rfds );

            timeval tv;
            tv.tv_sec  = wait_ms / 1000;
            tv.tv_usec = ( wait_ms % 1000 ) * 1000; // remember, these are usecs
            int ret = ::select( m_sockfd+1, &rfds, NULL, NULL, &tv);
            if ( ret == 0 )
            {
                // time limit expired
                cerr << "WARNING: Timeout reading from socket!" << endl;
                return false;
            }
            else if ( ret == SOCKET_ERROR )
            {
                // error
                int err = independent::sock_error();
                cerr << "Error waiting for data on read socket. err=" << err << endl;
                return false;
            }
        }

        size_t bytesleft = bufsz;
        char * pbuf = static_cast<char *>( buf );

	const bool log_performance_data = atoi(userSettings::obj().getValue("log_performance_data","0").c_str())>0;
	const string log_fn = "performance_data";
        fstream out( log_fn.c_str(), ios::app );

        timeval start_tv;
        independent::get_time( start_tv );
        simple_stats _stats;
        _stats.addnode( 0, start_tv );

        if ( log_performance_data && m_readRate )
        {
            out << "START TIME(read): [" << start_tv.tv_sec << "-" << start_tv.tv_usec << "], bytes=" << bufsz << "."  << endl;
        }

        while ( bytesleft > 0 )
        {
            size_t bytes_this_round = bytesleft;

            // We throttle by chunking the send and delaying between sends... the current
            // (cheap ass) implementation uses quarter second chunks.
            if ( m_writeRate > 0 )
                bytes_this_round = ( m_writeRate < 2000 ) ? 1000 : m_writeRate / 4;
            if ( bytes_this_round > bytesleft )
                bytes_this_round = bytesleft;

            // Do the read
            int bytesread = recv( m_sockfd, pbuf, bytes_this_round, 0 );
            if ( bytesread == -1 )
            {   // ERR
                // If we had an interrupt, try again...
                if ( errno == EINTR )
                    bytesread = 0;
                // Otherwise, we're done.
                else
                    return false;
            }
            else if ( bytesread == 0 )
            {
                // EOF
                break;
            }

            if ( log_performance_data && m_readRate )
            {
                out << "READ: read " << bytesread << " bytes." << endl;
            }

            bytesleft -= bytesread;
            pbuf      += bytesread;

            // If the monkeys want feedback, give them a banana
            if ( m_cb )
                (*m_cb)(bufsz - bytesleft, bufsz);

            // Do the throttle if necessary...
            timeval end_tv;
            independent::get_time( end_tv );
            while ( m_readRate && bufsz >= 1024 && _stats.addnode( bytesread, end_tv ) > m_readRate )
            {
                bytesread = 0; // accounted for...
                // lets hold up a quarter second...
                if ( log_performance_data )
                {
		    // The 10000 is to scale back the quantity of information (largely useless) that is
		    // generated by quarter second logging.  Its insane how much data that was before this
		    // millisecond check was in.
		    static unsigned long last= end_tv.tv_usec / 10000;
		    if ( end_tv.tv_usec / 10000 != last )
		    {
			out << "WAIT(read): [ " << end_tv.tv_sec << "-" << end_tv.tv_usec << " ], target_rate=" << m_readRate << ", current_rate=" << _stats.compute_rate() << endl;
			last = end_tv.tv_usec/10000;
		    }
                }
                static timeval quarter_sec = { 0, 250000 };
                select( 0, NULL, NULL, NULL, &quarter_sec );
                independent::get_time( end_tv );
            }
        }

	if ( ! log_performance_data ) unlink( log_fn.c_str() );

        return ( bytesleft == 0 );
    }

    //
    // writebuf() :
    //
    bool socket::writebuf( const void * buf, const size_t bufsz )
    {
        assert( is_connected() && buf && bufsz > 0 );

        size_t bytesleft = bufsz;
        const char * pbuf  = static_cast<const char *>( buf );

	const bool log_performance_data = atoi(userSettings::obj().getValue("log_performance_data","0").c_str())>0;
	const string log_fn = "performance_data";
        fstream out( log_fn.c_str(), ios::app );

        timeval start_tv;
        independent::get_time( start_tv );
        simple_stats _stats;
        _stats.addnode( 0, start_tv );

        if ( log_performance_data && m_writeRate )
        {
            out << "START TIME(write): [" << start_tv.tv_sec << "-" << start_tv.tv_usec << "], bytes=" << bufsz << "."  << endl;
        }

        while ( bytesleft > 0 )
        {
            size_t bytes_this_round = bytesleft;

            // We throttle by chunking the send and delaying between sends... the current
            // (cheap ass) implementation uses quarter second chunks.
            if ( m_writeRate > 0 )
                bytes_this_round = ( m_writeRate < 2000 ) ? 1000 : m_writeRate / 4;
            if ( bytes_this_round > bytesleft )
                bytes_this_round = bytesleft;

            // Do the send....
            int byteswritten = send( m_sockfd, pbuf, bytes_this_round, 0 );
            if ( byteswritten == -1 )
            {   // ERR
                // If we had an interrupt, try again...
                if ( errno == EINTR )
                    byteswritten = 0;
                // Otherwise, we're done.
                else
                    return false;
            }

            if ( log_performance_data && m_writeRate )
            {
                out << "WRITE: wrote " << byteswritten << " bytes." << endl;
            }

            bytesleft -= byteswritten;
            pbuf      += byteswritten;

            // If the monkeys want feedback, give them a banana
            if ( m_cb )
                (*m_cb)(bufsz - bytesleft, bufsz);

            // Do the throttle if necessary...
            timeval end_tv;
            independent::get_time( end_tv );
            while ( m_writeRate && bufsz >= 1024 && _stats.addnode( byteswritten, end_tv ) > m_writeRate )
            {
                byteswritten = 0; // accounted for...
                // lets hold up a quarter second...
                if ( log_performance_data )
                {
		    // The 10000 is to scale back the quantity of information (largely useless) that is
		    // generated by quarter second logging.  Its insane how much data that was before this
		    // millisecond check was in.
		    static unsigned long last= end_tv.tv_usec / 10000;
		    if ( end_tv.tv_usec / 10000 != last )
		    {
			out << "WAIT(write): [ " << end_tv.tv_sec << "-" << end_tv.tv_usec << " ], target_rate=" << m_writeRate << ", current_rate=" << _stats.compute_rate() << endl;
			last = end_tv.tv_usec/10000;
		    }
                }

                static timeval quarter_sec = { 0, 250000 };
                select( 0, NULL, NULL, NULL, &quarter_sec );
                independent::get_time( end_tv );
            }
        }

	if ( ! log_performance_data ) unlink( log_fn.c_str() );

        return ( bytesleft == 0 );
    }

    //
    // writebuf() :
    //
    //bool socket::writebuf( const void * buf, const size_t bufsz )
    // {
    //return basic_sleep( buf, bufsz );
    //}

    //
    // readline() : This is a cheap ass byte by byte version. I really should fix this.
    //
    bool socket::readline( std::string & str, const size_t max_len, unsigned int wait_ms )
    {
        assert( is_connected() );

        CHoldIntUntilDestruction _hold( m_readRate );

        str.erase();

        // Wait for input...
        if ( wait_ms )
        {
            fd_set rfds;
            FD_ZERO( &rfds );
            FD_SET( m_sockfd, &rfds );

            timeval tv;
            tv.tv_sec  = wait_ms / 1000;
            tv.tv_usec = ( wait_ms % 1000 ) * 1000; // remember, these are usecs
            int ret = ::select( m_sockfd+1, &rfds, NULL, NULL, &tv);
            if ( ret == 0 )
            {
                // time limit expired
                cerr << "Timeout reading from socket" << endl;
                return false;
            }
            else if ( ret == SOCKET_ERROR )
            {
                // error
                int err = independent::sock_error();
                cerr << "Error waiting for data on read socket. err=" << err << endl;
                return false;
            }
        }

        // optimize this later
        size_t bytesread = 0;
        char c = 0;
        while ( bytesread < max_len )
        {
            const int ret = readbuf( &c, 1, 0 );
            if ( ret == 1 )
            {
                // These characters can terminate a string as far as we are concerned
                if ( c == 0 || c == '\n' )
                    break;
                // These chars are meaningful to us
                else
                    str += c;
            }
            else
            {
                // We're done with this string...
                break;
            }
        }

        // Slice of tailing \r\n or \n or \r

        const int endidx = str.size() - 1;
        if ( endidx > 0 )
        {
            if ( str[ endidx - 1 ] == '\r' && str[ endidx ] == '\n' )
            {
                // cull the last two chars
                str = string( str.c_str(), str.size() - 2 );
            }
            else if ( str[ endidx ] == '\r' || str[ endidx ] == '\n' )
            {
                // cull the trailing doodad
                str = string( str.c_str(), str.size() - 1 );
            }
        }

        return ( str.size() > 0 );
    }

    //
    // writeline() :
    //
    bool socket::writeline( const std::string & str, const std::string & terminator )
    {
        assert( is_connected() && str.size() );

        bool bret =  writebuf( str.c_str(), str.size() );
        if ( terminator.size() )
        {
            bret &= writebuf( terminator.c_str(), terminator.size() );
        }

        return bret;
    }

    //
    // connect() :
    //
    bool socket::connect( const sockaddr & addr )
    {
        return connect( AF_INET, SOCK_STREAM, addr, 0 );
    }
    bool socket::connect( const int address_family, const int socket_type, const sockaddr & addr, const int protocol )
    {
        m_sockfd = ::socket( address_family, socket_type, protocol );
        if ( -1 == m_sockfd )
            return false;
        if ( -1 == ::connect( m_sockfd, addr.get_address(), addr.get_address_size() ) )
            return false;
        int keep_alive = 0;
        if ( 0 != ::setsockopt( m_sockfd, SOL_SOCKET, SO_KEEPALIVE|TCP_NODELAY, reinterpret_cast<const char *>( &keep_alive ), sizeof(keep_alive) ) < 0)
            return false;
        return true;
    }

    //
    // is_connected() :
    //
    bool socket::is_connected()
    const
    {
        return ( m_sockfd != -1 );
    }

    //
    // close() :
    //
    bool socket::close()
    {
        bool ret = true;
        assert( is_connected() );
        if ( m_sockfd != -1 )
        {
            if ( ::close( m_sockfd ) == -1 )
                ret = false;
            else
                m_sockfd = -1;
        }

        return ret;
    }


    ////////////////////////////////////////////////////////////
    //
    // listener
    //

    //
    // ctor/dtor
    //
    listener::~listener()
    {
        if ( is_connected() )
            close();
    }

    listener::listener( const sockaddr & listen_on )
    {
        m_sockfd = ::socket( AF_INET, SOCK_STREAM, 0 );
        if ( m_sockfd == -1 )
            throw runtime_error( "Failed to create listener socket" );

        if ( ::bind( m_sockfd, listen_on.get_address(), listen_on.get_address_size() ) == -1 )
            throw runtime_error( string( "Failed to listen on " ) + listen_on.get_verbose_address() );

        if ( ::listen( m_sockfd, 1024 ) == -1 )
            throw runtime_error( "Failed to start listening" );

        cerr << "Socket open and ready for business: " << listen_on.get_verbose_address() << endl;
    }

    //
    // is_connected() :
    //
    bool listener::is_connected() const
    {
        return ( m_sockfd != -1 );
    }

    //
    // close() :
    //
    bool listener::close()
    {
        bool ret = true;
        assert( is_connected() );
        if ( m_sockfd != -1 )
        {
            if ( ::close( m_sockfd ) == -1 )
                ret = false;
            else
                m_sockfd = -1;
        }

        return ret;
    }

    //
    // listen() :
    //
    bool listener::listen( socket & client_sock, sockaddr & client_addr )
    {
        sockaddr_in addr_clt;

        socklen_t cltaddrsz = 0;

        // Cool, this is one of those places where you need a space between the < and the name
        int connfd = ::accept( m_sockfd, reinterpret_cast< ::sockaddr*>( &addr_clt ) , &cltaddrsz );
        if ( connfd == -1 )
            return false;

        assert( cltaddrsz == sizeof( addr_clt ) );

        client_addr.set_address( reinterpret_cast< ::sockaddr*>( &addr_clt ) );
        assert( ! client_sock.is_connected() );
        client_sock.attach( connfd );
        return true;
    }
}
