// $Id: socket.h,v 1.4 2002/04/22 23:07:41 jjlupa Exp $
//
// ypost (Yet Another Poster)
// This code is copyright Jonathan Lupa, 2000. <jjlupa@jamdata.net>  To the
//  extent that it is released, it is released under the GNU General Public License
//  v2 or greater. The author retains all rights with respect to this code.
//
// If you have a desire to use this code outside of the GPL, drop me an email and
//  I would be happy to work with you under an alternate licence.  If you just want
//  to write code, then I would probably be happy to release it to you as under
//  the perl artistic license.

// sockets.h -- simple socket wrapping.

#ifndef core_sockets_h_
#define core_sockets_h_

#include <string>
#include <exception>

struct sockaddr;

namespace sock
{
    ////////////////////////////////////////////////////////////////////////////////
    //
    // AF indendent socket address.  Wouldn't it be cool if AF_INET6 actually worked with this code? :)
    //
    class sockaddr_imp;
    class sockaddr
    {
    public:
        //
        // Basicly wraps the socket naming... address name can be a dotted quad, hex string, DNS name, etc.
        //  service_name can be an IPv6 service name or a port number (or IPv4 service name[/etc/services] I think.
        //
        sockaddr();
        sockaddr( const std::string & address_name, const std::string & service_name );
        bool set_address( const std::string & address_name, const std::string & service_name );
        bool set_address( const ::sockaddr * addr_in );

        const ::sockaddr * get_address() const;
        // Important that you use this if AF_INET6 is even a shadow of a dream
        size_t             get_address_size() const;

        std::string get_verbose_address() const;
        bool is_valid() const;

    private:
        sockaddr_imp * m_pimp;
    };


    ////////////////////////////////////////////////////////////////////////////////
    //
    // socket i/o interface. all sockets provide these functions
    //

    typedef void (*progress_callback)(long, long);

    class socket_io
    {
    public:
        // Arbitrary way to send data in blobs over the connected socket.  No surprises here.
        //
        virtual bool readbuf( void * buf, const size_t bufsz, unsigned int wait_ms = 600000 ) = 0;
        virtual bool writebuf( const void * buf, const size_t bufsz ) = 0;

        // Basicly a string handling API... Note that readline will clip \r or \r\n, and that writeline appends
        // them by default.  You can keep writeline from appending them by sending a string of length 0...
        //
        virtual bool readline( std::string & str, const size_t max_len = 0, unsigned int wait_ms = 600000 ) = 0;
        virtual bool writeline( const std::string & str, const std::string & terminator = std::string( "\r\n" ) ) = 0;

        // A clean way to tell if we maybe shouldn't be calling the read/write commands
        //
        virtual bool is_connected() const = 0;

        // If cb is a valid pointer then you get feedback information about how far along we are in the current
        // read or write block.  Expect weirdness if you are looking for feedback on reading strings. I mean, think
        // about it.
        //
        virtual void set_progress_callback( progress_callback cb ) = 0;
    };


    ////////////////////////////////////////////////////////////////////////////////
    //
    // throttling interface. Set to 0 to disable throttling.
    //
    class throttleable
    {
    public:
        virtual void read_throttle( int bytes_per_second ) = 0;
        virtual void write_throttle( int bytes_per_second ) = 0;
    };


    ////////////////////////////////////////////////////////////////////////////////
    //
    // socket() : our socket implementation
    //
    class socket : public socket_io, public throttleable
    {
        // Construction
    public:
        socket();
        ~socket();

        // A quick note about this.  This construction passes ownership of the fd...
        // be careful about how its used.  Also consider fork implications
        void attach( int fd );

        // Identification
    public:
        static const std::string null_terminator; // = ""

        // Interface
    public:
        virtual bool readbuf( void * buf, const size_t bufsz, unsigned int wait_ms = 60000 );
        virtual bool writebuf( const void * buf, const size_t bufsz );

        virtual bool readline( std::string & str, const size_t max_len = 0, unsigned int wait_ms = 60000 );
        virtual bool writeline( const std::string & str, const std::string & terminator = std::string( "\r\n" ) );

        virtual void read_throttle( int bytes_per_second )
        {
            m_readRate = bytes_per_second;
        }
        virtual void write_throttle( int bytes_per_second )
        {
            m_writeRate = bytes_per_second;
        }

        bool connect( const sockaddr & addr );
        bool connect( int address_family, int socket_type, const sockaddr & addr, int protocol = 0 );

        virtual bool is_connected() const;
        virtual bool close();

        virtual void set_progress_callback( progress_callback cb )
        {
            m_cb = cb;
        }

        // Implementation
    private:
        bool basic_sleep( const void * buf, size_t bufsz, size_t timechunk_usec = 250000 /*1/4 sec*/ );
        bool rolling_sleep( const void * buf, const size_t bufsz, const size_t windowsize_usec );

        int m_sockfd;

        unsigned int m_readRate;
        unsigned int m_writeRate;

        progress_callback m_cb;
    };


    ////////////////////////////////////////////////////////////////////////////////
    //
    // listener : binds and waits
    //
    class listener
    {
        // Construction
    public:
        listener( const sockaddr & listen_on );
        ~listener();

        // Interface
    public:
        // Returns true if there is data pending on the socket. Returns false on socket close. If
        // successful, fills in the socket and address
        bool listen( socket & client_sock, sockaddr & client_addr );

        bool is_connected() const;
        bool close();
        // Implementation
    private:
        int m_sockfd;
    };
}

#endif

